package akhlys.fastdirsync;

import java.awt.Color;
import java.awt.Component;
import java.awt.Cursor;
import java.awt.Dimension;
import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;
import java.io.File;
import java.nio.file.Files;
import java.nio.file.Path;
import java.nio.file.Paths;
import java.util.List;

import javax.swing.BorderFactory;
import javax.swing.Box;
import javax.swing.BoxLayout;
import javax.swing.JButton;
import javax.swing.JFileChooser;
import javax.swing.JFrame;
import javax.swing.JLabel;
import javax.swing.JOptionPane;
import javax.swing.JPanel;
import javax.swing.JProgressBar;
import javax.swing.JScrollPane;
import javax.swing.JTable;
import javax.swing.JTextField;
import javax.swing.SwingUtilities;
import javax.swing.table.DefaultTableCellRenderer;

/**
 * Syncs directories using the file size to compare files
 * @author akhlys
 *
 */
public class FastDirSync extends JFrame implements ActionListener {
	private static final long serialVersionUID = -6707689228592657038L;
	
	private static double VERSION = 1.0;

	/**
	 * @param args
	 */
	public static void main(String[] args) {
		SwingUtilities.invokeLater(new Runnable() {
			public void run() {
				FastDirSync app = new FastDirSync();
				app.setVisible(true);
			}
		});
	}
	
	// normal members
	NicelyCancelableThread<?, Void> currentTask = null;
	protected Path sourceOfLastCompare = null;
	protected Path targetOfLastCompare = null;
	
	// swing components
	JPanel mainPanel = new JPanel();
	JPanel sourceDirPanel = new JPanel();
	JPanel targetDirPanel = new JPanel();
	JPanel compareButtonPanel = new JPanel();
	JPanel syncButtonPanel = new JPanel();
	JPanel progressPanel = new JPanel();
	
	JLabel sourceDirLabel = new JLabel("Source Directory");
	JLabel targetDirLabel = new JLabel("Target Directory");

	JTextField sourceDirPath = new JTextField();
	JTextField targetDirPath = new JTextField();
	
	JButton sourceDirButton = new JButton("Choose..");
	JButton targetDirButton = new JButton("Choose..");
	JButton compareButton = new JButton("Compare!");
	JButton syncButton = new JButton("Synchronize!");
	JButton abortButton = new JButton("Abort");
	
	JScrollPane resultScrollPane = null;
	JTable resultTable = null;
	
	JProgressBar progressBar = new JProgressBar(0, 100);
	
	/**
	 * ctor builds the GUI
	 */
	public FastDirSync() {
		this.mainPanel.setLayout(new BoxLayout(this.mainPanel, BoxLayout.Y_AXIS));
		this.sourceDirPanel.setLayout(new BoxLayout(this.sourceDirPanel, BoxLayout.X_AXIS));
		this.targetDirPanel.setLayout(new BoxLayout(this.targetDirPanel, BoxLayout.X_AXIS));
		this.compareButtonPanel.setLayout(new BoxLayout(this.compareButtonPanel, BoxLayout.X_AXIS));
		this.syncButtonPanel.setLayout(new BoxLayout(this.syncButtonPanel, BoxLayout.X_AXIS));
		this.progressPanel.setLayout(new BoxLayout(this.progressPanel, BoxLayout.X_AXIS));
		
		this.mainPanel.setBorder(BorderFactory.createEmptyBorder(5, 5, 5, 5));
		
		// source file
		this.sourceDirButton.addActionListener(this);
		this.sourceDirPanel.add(this.sourceDirLabel);
		this.sourceDirPanel.add(Box.createHorizontalStrut(10));
		this.sourceDirPanel.add(this.sourceDirPath);
		this.sourceDirPanel.add(Box.createHorizontalStrut(10));
		this.sourceDirPanel.add(this.sourceDirButton);
		this.sourceDirPath.setMaximumSize(new Dimension(Integer.MAX_VALUE, this.sourceDirButton.getPreferredSize().height) );
		this.mainPanel.add(this.sourceDirPanel);
		this.mainPanel.add(Box.createVerticalStrut(10));

		// target file
		this.targetDirButton.addActionListener(this);
		this.targetDirPanel.add(this.targetDirLabel);
		this.targetDirPanel.add(Box.createHorizontalStrut(10));
		this.targetDirPanel.add(this.targetDirPath);
		this.targetDirPanel.add(Box.createHorizontalStrut(10));
		this.targetDirPanel.add(this.targetDirButton);
		this.targetDirPath.setMaximumSize(new Dimension(Integer.MAX_VALUE, this.targetDirButton.getPreferredSize().height) );
		this.mainPanel.add(this.targetDirPanel);
		this.mainPanel.add(Box.createVerticalStrut(10));
		
		// quick hack to align the labels properly while still using the sumple BoxLayout
		int dirLabelWidth = Math.max(this.sourceDirLabel.getPreferredSize().width, this.targetDirLabel.getPreferredSize().width);
		this.sourceDirLabel.setPreferredSize(new Dimension(dirLabelWidth, this.sourceDirLabel.getPreferredSize().height));
		this.targetDirLabel.setPreferredSize(new Dimension(dirLabelWidth, this.targetDirLabel.getPreferredSize().height));
		
		// compare button
		this.compareButton.addActionListener(this);
		this.compareButtonPanel.add(this.compareButton);
		this.compareButtonPanel.add(Box.createGlue());
		this.mainPanel.add(this.compareButtonPanel);
		this.mainPanel.add(Box.createVerticalStrut(10));
		
		// table sceleton
		//CheckboxTableModel
		this.resultTable = new JTable(new FileDifferencesTableModel());
		this.resultTable.setDefaultRenderer(Object.class, new ColoredTableCellRenderer());
		//this.resultTable.setDefaultRenderer(Boolean.class, new ColoredTableCellRenderer());
		this.resultTable.setAutoCreateRowSorter(true);
		this.resultTable.setAutoResizeMode(JTable.AUTO_RESIZE_OFF);
		this.resultTable.getColumnModel().getColumn(0).setMaxWidth(30);
		this.resultTable.getColumnModel().getColumn(0).setMinWidth(30);
		this.resultTable.getColumnModel().getColumn(1).setPreferredWidth(200);
		this.resultTable.getColumnModel().getColumn(2).setPreferredWidth(342);
		this.resultTable.getColumnModel().getColumn(3).setPreferredWidth(170);
		
		this.resultScrollPane = new JScrollPane(this.resultTable);
		this.mainPanel.add(this.resultScrollPane);
		this.mainPanel.add(Box.createVerticalStrut(10));
		
		// sync button
		this.syncButton.addActionListener(this);
		this.syncButton.setEnabled(false);
		this.syncButtonPanel.add(this.syncButton);
		this.syncButtonPanel.add(Box.createGlue());
		this.mainPanel.add(this.syncButtonPanel);
		this.mainPanel.add(Box.createVerticalStrut(10));
		
		// progress bar
		this.abortButton.addActionListener(this);
		this.abortButton.setEnabled(false);
		this.progressBar.setStringPainted(true);
		this.progressBar.setString(""); // just setting stringpainted to false results in smaller progressbars
		this.progressPanel.add(this.progressBar);
		this.progressPanel.add(Box.createHorizontalStrut(10));
		this.progressPanel.add(this.abortButton);
		this.mainPanel.add(this.progressPanel);
		
		// general JFrame stuff
		this.add(this.mainPanel);
		this.setSize(755, 480);
		this.setDefaultCloseOperation(JFrame.EXIT_ON_CLOSE);
		this.setTitle("Akhlys' FastDirSync v" + VERSION);
		this.setLocationRelativeTo(null);
	}

	@Override
	public void actionPerformed(ActionEvent arg0) {
		Object source = arg0.getSource();
		
		if (source == this.sourceDirButton) {
			JFileChooser chooser = new JFileChooser(this.sourceDirPath.getText());
			chooser.setFileSelectionMode(JFileChooser.DIRECTORIES_ONLY);
			
			//chooser.setMultiSelectionEnabled(false);
			int returnCode = chooser.showOpenDialog(this);
			if (returnCode != JFileChooser.APPROVE_OPTION) return;
			File f = chooser.getSelectedFile();
			if (f.exists() && f.isDirectory()) this.sourceDirPath.setText(f.getAbsolutePath());
		} else if (source == this.targetDirButton) {
			JFileChooser chooser = new JFileChooser(this.targetDirPath.getText());
			chooser.setFileSelectionMode(JFileChooser.DIRECTORIES_ONLY);
			
			//chooser.setMultiSelectionEnabled(false);
			int returnCode = chooser.showOpenDialog(this);
			if (returnCode != JFileChooser.APPROVE_OPTION) return;
			File f = chooser.getSelectedFile();
			if (f.exists() && f.isDirectory()) this.targetDirPath.setText(f.getAbsolutePath());
		} else if (source == this.compareButton) {
			compareDirectories();
		} else if (source == this.syncButton) {
			//System.out.println(this.resultTable.getDefaultRenderer(Boolean.class));
			sync();
		} else if (source == this.abortButton) {
			if (this.currentTask != null) {
				this.currentTask.abort();
			}
			this.abortButton.setEnabled(false);
		} else {
			System.out.println("Error in actionPerformed: source object is unknown obj " + source.toString());
		}
	}
	
	private void compareDirectories() {
		Path sourceRootDir;
		Path targetRootDir;
		sourceRootDir = Paths.get(this.sourceDirPath.getText()).normalize(); // toRealPath() would normalize and check for existance but does so by way of exception, this here is cleaner
		targetRootDir = Paths.get(this.targetDirPath.getText()).normalize();
		if (!Files.exists(sourceRootDir) || !Files.isDirectory(sourceRootDir) || !Files.exists(targetRootDir) || !Files.isDirectory(targetRootDir)
				|| isChildOf(sourceRootDir, targetRootDir) || isChildOf(targetRootDir, sourceRootDir)) {
			JOptionPane.showMessageDialog(this, "Source or target directory path not valid.", "Error", JOptionPane.ERROR_MESSAGE);
			return;
		}
		
		lockGUIForTask(true);
		
		this.progressBar.setString("");
		this.progressBar.setIndeterminate(true);
		
		this.currentTask = new DifferenceFinderThread(this, sourceRootDir, targetRootDir);
		this.currentTask.execute();
	}
	
	private void sync() {
		Path sourceRootDir = this.sourceOfLastCompare;
		Path targetRootDir = this.targetOfLastCompare;
		if (sourceRootDir == null || targetRootDir == null) {
			JOptionPane.showMessageDialog(this, "Source or target directory path not valid.", "Error", JOptionPane.ERROR_MESSAGE);
			return;
		}
		
		lockGUIForTask(true);
		
		this.progressBar.setString("");
		this.progressBar.setIndeterminate(true);
		
		List<FileDifference> differences = ((FileDifferencesTableModel)this.resultTable.getModel()).getSelectedFileDifferences();
		
		this.currentTask = new SynchronizationThread(this, sourceRootDir, targetRootDir, differences);
		this.currentTask.execute();
	}
	
	private boolean isChildOf(Path maybeChild, Path possibleParent) {
		Path parent = possibleParent.normalize();
		if (!Files.exists(parent) || !Files.isDirectory(parent)) {
			// this cannot possibly be the parent
			return false;
		}
		
		Path child = maybeChild.normalize();
		while (child != null) {
			if (child.equals(parent)) {
				return true;
			}
			child = child.getParent();
		}
		// No match found, and we've hit the root directory
		return false;
	}

	
	/**
	 * sets everything but the abort button to disabled and sets a waiting cursor if lock = true
	 * @param lock
	 */
	public void lockGUIForTask(boolean lock) {
		this.sourceDirButton.setEnabled(!lock);
		this.sourceDirPath.setEnabled(!lock);
		this.targetDirButton.setEnabled(!lock);
		this.targetDirPath.setEnabled(!lock);
		this.compareButton.setEnabled(!lock);
		this.syncButton.setEnabled(!lock);
		this.resultTable.setEnabled(!lock);
		
		this.abortButton.setEnabled(lock);
		
		if (lock) {
			this.mainPanel.setCursor(Cursor.getPredefinedCursor(Cursor.WAIT_CURSOR));
		} else {
			this.mainPanel.setCursor(Cursor.getPredefinedCursor(Cursor.DEFAULT_CURSOR));
		}
		this.progressBar.setString("");
	}
}

class ColoredTableCellRenderer extends DefaultTableCellRenderer {
	private static final long serialVersionUID = -6122870807947271117L;

	@Override
	public Component getTableCellRendererComponent(JTable table, Object value, boolean isSelected, boolean hasFocus, int row, int column) {
		Component comp = super.getTableCellRendererComponent(table, value, isSelected, hasFocus, row, column);
		if (comp == null) return null;
		
		int modelRow = table.convertRowIndexToModel(row); // row = row from view, not model
		FileDiffType diff = ((FileDifferencesTableModel)table.getModel()).getFileDifferenceAt(modelRow).diffType;
		
		switch (diff) {
			case CHANGED:
				comp.setBackground(Color.orange);
				break;
			case CHANGED_TYPE_D2F:
			case CHANGED_TYPE_F2D:
				comp.setBackground(Color.YELLOW);
				break;
			case DELETED:
				comp.setBackground(Color.pink);
				break;
			case NEW:
				comp.setBackground(Color.green);
				break;
		}
		
		return comp;
	}
}