package akhlys.fastdirsync;

/**
 * A enum for the different types of file differences
 * @author akhlys
 *
 */
public enum FileDiffType {
	/** A file/dir in the source directory is not in the target directory **/
	NEW ("NEW: File/Directory is in source but not in Target", "Copy source file to target"),
	/** A file/dir in the target directory is not in the source directory **/
	DELETED ("DELETED: File/Directory is in target but not in source", "Remove file from target"),
	/** A file is in both dirs but different **/
	CHANGED ("CHANGED: File exists in both, but was changed", "Replace file on target"),
	/** A file is a directory on target **/
	CHANGED_TYPE_F2D ("CHANGED TYPE: Same file name exists on target, but it's a directory", "Replace dir on target with source file"),
	/** A directory is a file on target **/
	CHANGED_TYPE_D2F ("CHANGED TYPE: Same directory name exists on target, but it's a file", "Replace file on target with source directory");
	
	private final String stringRepresentation;
	private final String requiredAction;
	FileDiffType(String stringrep, String requiredAction) {
        this.stringRepresentation = stringrep;
        this.requiredAction = requiredAction;
    }

	@Override
	public String toString() {
		return this.stringRepresentation;
	};
	
	/**
	 * Returns a description of the action needed to resolfe the diff
	 * @return the description 
	 */
	public String getRequiredAction() {
		return this.requiredAction;
	}
}
